@@ -1,16 +1,11 @@
 #!/usr/bin/env python
 import signal
 
-import sys, os, datetime, time, copy, json, traceback, csv, cPickle, string
+import sys, os, datetime, time, json, traceback, csv, cPickle
 import multiprocessing, multiprocessing.queues, subprocess, re, ctypes
 from optparse import OptionParser
-
-try:
-    import rethinkdb as r
-except ImportError:
-    print "The RethinkDB python driver is required to use this command."
-    print "Please install the driver via `pip install rethinkdb`."
-    exit(1)
+from ._backup import *
+import rethinkdb as r
 
 info = "'rethinkdb import` loads data into a RethinkDB cluster"
 usage = "\
@@ -30,6 +25,8 @@ def print_import_help():
     print "  -a [ --auth ] AUTH_KEY           authorization key for rethinkdb clients"
     print "  --clients NUM_CLIENTS            the number of client connections to use (defaults"
     print "                                   to 8)"
+    print "  --hard-durability                use hard durability writes (slower, but less memory"
+    print "                                   consumption on the server)"
     print "  --force                          import data even if a table already exists, and"
     print "                                   overwrite duplicate primary keys"
     print "  --fields                         limit which fields to use when importing one table"
@@ -80,6 +77,7 @@ def parse_options():
     parser.add_option("-a", "--auth", dest="auth_key", metavar="AUTHKEY", default="", type="string")
     parser.add_option("--fields", dest="fields", metavar="FIELD,FIELD...", default=None, type="string")
     parser.add_option("--clients", dest="clients", metavar="NUM_CLIENTS", default=8, type="int")
+    parser.add_option("--hard-durability", dest="hard", action="store_true", default=False)
     parser.add_option("--force", dest="force", action="store_true", default=False)
     parser.add_option("--debug", dest="debug", action="store_true", default=False)
 
@@ -109,18 +107,14 @@ def parse_options():
     res = { }
 
     # Verify valid host:port --connect option
-    host_port = options.host.split(":")
-    if len(host_port) == 1:
-        host_port = (host_port[0], "28015") # If just a host, use the default port
-    if len(host_port) != 2:
-        raise RuntimeError("Error: Invalid 'host:port' format: %s" % options.host)
-    (res["host"], res["port"]) = host_port
+    (res["host"], res["port"]) = parse_connect_option(options.host)
 
     if options.clients < 1:
         raise RuntimeError("Error: --client option too low, must have at least one client connection")
 
     res["auth_key"] = options.auth_key
     res["clients"] = options.clients
+    res["durability"] = "hard" if options.hard else "soft"
     res["force"] = options.force
     res["debug"] = options.debug
 
@@ -154,23 +148,12 @@ def parse_options():
             raise RuntimeError("Error: Directory to import does not exist: %d" % res["directory"])
 
         # Verify valid --import options
-        res["dbs"] = []
-        res["tables"] = []
-        for item in options.tables:
-            if not all(c in string.ascii_letters + string.digits + "._" for c in item):
-                raise RuntimeError("Error: Invalid 'db' or 'db.table' name: %s" % item)
-            db_table = item.split(".")
-            if len(db_table) == 1:
-                res["dbs"].append(db_table[0])
-            elif len(db_table) == 2:
-                res["tables"].append(tuple(db_table))
-            else:
-                raise RuntimeError("Error: Invalid 'db' or 'db.table' format: %s" % item)
+        res["db_tables"] = parse_db_table_options(options.tables)
 
         # Parse fields
         if options.fields is None:
             res["fields"] = None
-        elif len(res["dbs"]) != 0 or len(res["tables"] != 1):
+        elif len(res["db_tables"]) != 1 or res["db_tables"][0][1] is None:
             raise RuntimeError("Error: Can only use the --fields option when importing a single table")
         else:
             res["fields"] = options.fields.split(",")
@@ -203,12 +186,9 @@ def parse_options():
         # Verify valid --table option
         if options.import_table is None:
             raise RuntimeError("Error: Must specify a destination table to import into using the --table option")
-        if not all(c in string.ascii_letters + string.digits + "._" for c in options.import_table):
-            raise RuntimeError("Error: Invalid 'db' or 'db.table' name: %s" % options.import_table)
-        db_table = options.import_table.split(".")
-        if len(db_table) != 2:
-            raise RuntimeError("Error: Invalid 'db.table' format: %s" % db_table)
-        res["import_db_table"] = db_table
+        res["import_db_table"] = parse_db_table(options.import_table)
+        if res["import_db_table"][1] is None:
+            raise RuntimeError("Error: Invalid 'db.table' format: %s" % options.import_table)
 
         # Parse fields
         if options.fields is None:
@@ -249,23 +229,55 @@ def parse_options():
 
     return res
 
-# This is run for each client requested, and accepts tasks from the reader processes
-def client_process(host, port, auth_key, task_queue, error_queue, use_upsert):
-    try:
-        conn = r.connect(host, port, auth_key=auth_key)
-        while True:
-            task = task_queue.get()
-            if len(task) == 3:
-                # Unpickle objects (TODO: super inefficient, would be nice if we could pass down json)
-                objs = [cPickle.loads(obj) for obj in task[2]]
-                res = r.db(task[0]).table(task[1]).insert(objs, durability="soft", upsert=use_upsert).run(conn)
-                if res["errors"] > 0:
-                    raise RuntimeError("Error when importing into table '%s.%s': %s" %
-                                       (task[0], task[1], res["first_error"]))
+# This is called through rdb_call_wrapper so reattempts can be tried as long as progress
+# is being made, but connection errors occur.  We save a failed task in the progress object
+# so it can be resumed later on a new connection.
+def import_from_queue(progress, conn, task_queue, error_queue, use_upsert, durability):
+    if progress[0] is None:
+        progress[0] = 0
+        progress.append(None)
+    elif not use_upsert:
+        # We were interrupted and it's not ok to overwrite rows, check that the batch either:
+        # a) does not exist on the server
+        # b) is exactly the same on the server
+        task = progress[1]
+        pkey = r.db(task[0]).table(task[1]).info().run(conn)["primary_key"]
+        for i in reversed(xrange(len(task[2]))):
+            obj = cPickle.loads(task[2][i])
+            if pkey not in obj:
+                raise RuntimeError("Connection error while importing.  Current row has no specified primary key, so cannot guarantee absence of duplicates")
+            row = r.db(task[0]).table(task[1]).get(obj[pkey]).run(conn)
+            if row == obj:
+                progress[0] += 1
+                del task[2][i]
             else:
-                break
-    except (r.RqlError, r.RqlDriverError) as ex:
-        error_queue.put((RuntimeError, RuntimeError(ex.message), traceback.extract_tb(sys.exc_info()[2])))
+                raise RuntimeError("Duplicate primary key `%s`:\n%s\n%s" % (pkey, str(obj), str(row)))
+
+    task = task_queue.get() if progress[1] is None else progress[1]
+    while len(task) == 3:
+        try:
+            # Unpickle objects (TODO: super inefficient, would be nice if we could pass down json)
+            objs = [cPickle.loads(obj) for obj in task[2]]
+            res = r.db(task[0]).table(task[1]).insert(objs, durability=durability, upsert=use_upsert).run(conn)
+        except:
+            progress[1] = task
+            raise
+
+        if res["errors"] > 0:
+            raise RuntimeError("Error when importing into table '%s.%s': %s" %
+                               (task[0], task[1], res["first_error"]))
+
+        progress[0] += len(objs)
+        task = task_queue.get()
+    return progress[0]
+
+# This is run for each client requested, and accepts tasks from the reader processes
+def client_process(host, port, auth_key, task_queue, error_queue, rows_written, use_upsert, durability):
+    try:
+        conn_fn = lambda: r.connect(host, port, auth_key=auth_key)
+        res = rdb_call_wrapper(conn_fn, "import", import_from_queue, task_queue, error_queue, use_upsert, durability)
+        with rows_written.get_lock():
+            rows_written.value += res
     except:
         ex_type, ex_class, tb = sys.exc_info()
         error_queue.put((ex_type, ex_class, traceback.extract_tb(tb)))
@@ -335,7 +347,6 @@ def read_json_array(json_data, file_in, callback, progress_info):
 
             (obj, offset) = decoder.raw_decode(json_data, idx=offset)
             callback(obj)
-            progress_info[2].value += 1
 
             # Read past whitespace to the next record
             file_offset += offset
@@ -361,7 +372,7 @@ def read_json_array(json_data, file_in, callback, progress_info):
     json_data += file_in.read()
     return json_data[offset + 1:]
 
-def json_reader(task_queue, filename, db, table, primary_key, fields, progress_info, exit_event):
+def json_reader(task_queue, filename, db, table, fields, progress_info, exit_event):
     object_buffers = []
     buffer_sizes = []
 
@@ -380,7 +391,6 @@ def json_reader(task_queue, filename, db, table, primary_key, fields, progress_i
             json_data = read_json_array(json_data[offset + 1:], file_in, callback, progress_info)
         elif json_data[offset] == "{":
             json_data = read_json_single_object(json_data[offset:], file_in, callback)
-            progress_info[2].value = 1
         else:
             raise RuntimeError("Error: JSON format not recognized - file does not begin with an object or array")
 
@@ -395,7 +405,7 @@ def json_reader(task_queue, filename, db, table, primary_key, fields, progress_i
     if len(object_buffers) > 0:
         task_queue.put((db, table, object_buffers))
 
-def csv_reader(task_queue, filename, db, table, primary_key, options, progress_info, exit_event):
+def csv_reader(task_queue, filename, db, table, options, progress_info, exit_event):
     object_buffers = []
     buffer_sizes = []
 
@@ -434,26 +444,29 @@ def csv_reader(task_queue, filename, db, table, primary_key, options, progress_i
                 if len(obj[key]) == 0:
                     del obj[key]
             object_callback(obj, db, table, task_queue, object_buffers, buffer_sizes, options["fields"], exit_event)
-            progress_info[2].value += 1
 
     if len(object_buffers) > 0:
         task_queue.put((db, table, object_buffers))
 
+# This function is called through rdb_call_wrapper, which will reattempt if a connection
+# error occurs.  Progress is not used as this will either succeed or fail.
+def create_table(progress, conn, db, table, pkey):
+    if table not in r.db(db).table_list().run(conn):
+        r.db(db).table_create(table, primary_key=pkey).run(conn)
+
 def table_reader(options, file_info, task_queue, error_queue, progress_info, exit_event):
     try:
         db = file_info["db"]
         table = file_info["table"]
         primary_key = file_info["info"]["primary_key"]
-        conn = r.connect(options["host"], options["port"], auth_key=options["auth_key"])
 
-        if table not in r.db(db).table_list().run(conn):
-            r.db(db).table_create(table, primary_key=primary_key).run(conn)
+        conn_fn = lambda: r.connect(options["host"], options["port"], auth_key=options["auth_key"])
+        rdb_call_wrapper(conn_fn, "create table", create_table, db, table, primary_key)
 
         if file_info["format"] == "json":
             json_reader(task_queue,
                         file_info["file"],
                         db, table,
-                        primary_key,
                         options["fields"],
                         progress_info,
                         exit_event)
@@ -461,14 +474,11 @@ def table_reader(options, file_info, task_queue, error_queue, progress_info, exi
             csv_reader(task_queue,
                        file_info["file"],
                        db, table,
-                       primary_key,
                        options,
                        progress_info,
                        exit_event)
         else:
             raise RuntimeError("Error: Unknown file format specified")
-    except (r.RqlError, r.RqlDriverError) as ex:
-        error_queue.put((RuntimeError, RuntimeError(ex.message), traceback.extract_tb(sys.exc_info()[2])))
     except InterruptedError:
         pass # Don't save interrupted errors, they are side-effects
     except:
@@ -496,7 +506,7 @@ def print_progress(ratio):
 
 def update_progress(progress_info):
     lowest_completion = 1.0
-    for (current, max_count, rows) in progress_info:
+    for (current, max_count) in progress_info:
         curr_val = current.value
         max_val = max_count.value
         if curr_val < 0:
@@ -523,6 +533,7 @@ def spawn_import_clients(options, files_info):
 
     try:
         progress_info = [ ]
+        rows_written = multiprocessing.Value(ctypes.c_longlong, 0)
 
         for i in range(options["clients"]):
             client_procs.append(multiprocessing.Process(target=client_process,
@@ -531,13 +542,14 @@ def spawn_import_clients(options, files_info):
                                                               options["auth_key"],
                                                               task_queue,
                                                               error_queue,
-                                                              options["force"])))
+                                                              rows_written,
+                                                              options["force"],
+                                                              options["durability"])))
             client_procs[-1].start()
 
         for file_info in files_info:
             progress_info.append((multiprocessing.Value(ctypes.c_longlong, -1), # Current lines/bytes processed
-                                  multiprocessing.Value(ctypes.c_longlong, 0), # Total lines/bytes to process
-                                  multiprocessing.Value(ctypes.c_longlong, 0))) # Total rows processed
+                                  multiprocessing.Value(ctypes.c_longlong, 0))) # Total lines/bytes to process
             reader_procs.append(multiprocessing.Process(target=table_reader,
                                                         args=(options,
                                                               file_info,
@@ -569,13 +581,13 @@ def spawn_import_clients(options, files_info):
         if error_queue.empty() and not interrupt_event.is_set():
             print_progress(1.0)
 
-        # Continue past the progress output line
         def plural(num, text):
             return "%d %s%s" % (num, text, "" if num == 1 else "s")
 
+        # Continue past the progress output line
         print ""
-        print "%s imported in %s" % (plural(sum([info[2].value for info in progress_info]), "row"),
-                                       plural(len(files_info), "table"))
+        print "%s imported in %s" % (plural(rows_written.value, "row"),
+                                     plural(len(files_info), "table"))
     finally:
         signal.signal(signal.SIGINT, signal.SIG_DFL)
 
@@ -596,16 +608,17 @@ def spawn_import_clients(options, files_info):
                 print >> sys.stderr, "In file: %s" % (error[3])
         raise RuntimeError("Errors occurred during import")
 
-def get_import_info_for_file(filename, db_filter, table_filter):
+def get_import_info_for_file(filename, db_table_filter):
     file_info = { }
     file_info["file"] = filename
     file_info["format"] = os.path.split(filename)[1].split(".")[-1]
     file_info["db"] = os.path.split(os.path.split(filename)[0])[1]
     file_info["table"] = os.path.split(filename)[1].split(".")[0]
 
-    if len(db_filter) > 0 or len(table_filter) > 0:
-        if file_info["db"] not in db_filter and (file_info["db"], file_info["table"]) not in table_filter:
-            return None
+    if len(db_table_filter) > 0:
+        if (file_info["db"], None) not in db_table_filter:
+            if (file_info["db"], file_info["table"]) not in db_table_filter:
+                return None
 
     info_filepath = os.path.join(os.path.split(filename)[0], file_info["table"] + ".info")
     with open(info_filepath, "r") as info_file:
@@ -613,10 +626,32 @@ def get_import_info_for_file(filename, db_filter, table_filter):
 
     return file_info
 
+def tables_check(progress, conn, files_info, force):
+    # Ensure that all needed databases exist and tables don't
+    db_list = r.db_list().run(conn)
+    for db in set([file_info["db"] for file_info in files_info]):
+        if db not in db_list:
+            r.db_create(db).run(conn)
+
+    # Ensure that all tables do not exist (unless --forced)
+    already_exist = []
+    for file_info in files_info:
+        table = file_info["table"]
+        db = file_info["db"]
+        if table in r.db(db).table_list().run(conn):
+            if not force:
+                already_exist.append("%s.%s" % (db, table))
+
+            extant_pkey = r.db(db).table(table).info().run(conn)["primary_key"]
+            if file_info["info"]["primary_key"] != extant_pkey:
+                raise RuntimeError("Error: Table '%s.%s' already exists with a different primary key" % (db, table))
+
+    return already_exist
+
 def import_directory(options):
     # Scan for all files, make sure no duplicated tables with different formats
     dbs = False
-    db_filter = set([db_table[0] for db_table in options["tables"]]) | set(options["dbs"])
+    db_filter = set([db_table[0] for db_table in options["db_tables"]])
     files_to_import = []
     files_ignored = []
     for (root, dirs, files) in os.walk(options["directory"]):
@@ -625,7 +660,7 @@ def import_directory(options):
             # The first iteration through should be the top-level directory, which contains the db folders
             dbs = True
             if len(db_filter) > 0:
-                for i in range(len(dirs)):
+                for i in reversed(xrange(len(dirs))):
                     if dirs[i] not in db_filter:
                         del dirs[i]
         else:
@@ -646,7 +681,7 @@ def import_directory(options):
     # For each table to import collect: file, format, db, table, info
     files_info = []
     for filename in files_to_import:
-        res = get_import_info_for_file(filename, options["dbs"], options["tables"])
+        res = get_import_info_for_file(filename, options["db_tables"])
         if res is not None:
             files_info.append(res)
 
@@ -660,32 +695,11 @@ def import_directory(options):
 
         db_tables.add((file_info["db"], file_info["table"]))
 
-    # Ensure that all needed databases exist and tables don't
-    try:
-        conn = r.connect(options["host"], options["port"], auth_key=options["auth_key"])
-    except (r.RqlError, r.RqlDriverError) as ex:
-        raise RuntimeError(ex.message)
-
-    db_list = r.db_list().run(conn)
-    for db in set([file_info["db"] for file_info in files_info]):
-        if db not in db_list:
-            r.db_create(db).run(conn)
-
-    # Ensure that all tables do not exist (unless --forced)
-    already_exist = []
-    for file_info in files_info:
-        table = file_info["table"]
-        db = file_info["db"]
-        if table in r.db(db).table_list().run(conn):
-            if not options["force"]:
-                already_exist.append("%s.%s" % (db, table))
-
-            extant_primary_key = r.db(db).table(table).info().run(conn)["primary_key"]
-            if file_info["info"]["primary_key"] != extant_primary_key:
-                raise RuntimeError("Error: Table '%s.%s' already exists with a different primary key" % (db, table))
+    conn_fn = lambda: r.connect(options["host"], options["port"], auth_key=options["auth_key"])
+    already_exist = rdb_call_wrapper(conn_fn, "tables check", tables_check, files_info, options["force"])
 
     if len(already_exist) == 1:
-        raise RuntimeError("Error: Table '%s.%s' already exists, run with --force to import into the existing table" % (db, table))
+        raise RuntimeError("Error: Table '%s' already exists, run with --force to import into the existing table" % already_exist[0])
     elif len(already_exist) > 1:
         already_exist.sort()
         extant_tables = "\n  ".join(already_exist)
@@ -701,34 +715,34 @@ def import_directory(options):
 
     spawn_import_clients(options, files_info)
 
-def import_file(options):
-    db = options["import_db_table"][0]
-    table = options["import_db_table"][1]
-    primary_key = options["primary_key"]
-
-    try:
-        conn = r.connect(options["host"], options["port"], auth_key=options["auth_key"])
-    except (r.RqlError, r.RqlDriverError) as ex:
-        raise RuntimeError(ex.message)
-
-    # Ensure that the database and table exist
+def table_check(progress, conn, db, table, pkey, force):
     if db not in r.db_list().run(conn):
         r.db_create(db).run(conn)
 
     if table in r.db(db).table_list().run(conn):
-        if not options["force"]:
+        if not force:
             raise RuntimeError("Error: Table already exists, run with --force if you want to import into the existing table")
 
-        extant_primary_key = r.db(db).table(table).info().run(conn)["primary_key"]
-        if primary_key is not None and primary_key != extant_primary_key:
+        extant_pkey = r.db(db).table(table).info().run(conn)["primary_key"]
+        if pkey is not None and pkey != extant_pkey:
             raise RuntimeError("Error: Table already exists with a different primary key")
-        primary_key = extant_primary_key
+        pkey = extant_pkey
     else:
-        if primary_key is None:
+        if pkey is None:
             print "no primary key specified, using default primary key when creating table"
             r.db(db).table_create(table).run(conn)
         else:
-            r.db(db).table_create(table, primary_key=primary_key).run(conn)
+            r.db(db).table_create(table, primary_key=pkey).run(conn)
+    return pkey
+
+def import_file(options):
+    db = options["import_db_table"][0]
+    table = options["import_db_table"][1]
+    pkey = options["primary_key"]
+
+    # Ensure that the database and table exist with the right primary key
+    conn_fn = lambda: r.connect(options["host"], options["port"], auth_key=options["auth_key"])
+    pkey = rdb_call_wrapper(conn_fn, "table check", table_check, db, table, pkey, options["force"])
 
     # Make this up so we can use the same interface as with an import directory
     file_info = {}
@@ -736,7 +750,7 @@ def import_file(options):
     file_info["format"] = options["import_format"]
     file_info["db"] = db
     file_info["table"] = table
-    file_info["info"] = { "primary_key": primary_key }
+    file_info["info"] = { "primary_key": pkey }
 
     spawn_import_clients(options, [file_info])
 
