@@ -1,30 +1,54 @@
 # Copyright 2010-2012 RethinkDB, all rights reserved.
 
-__all__ = ['connect', 'Connection', 'Cursor', 'protobuf_implementation']
+__all__ = ['connect', 'Connection', 'Cursor']
 
 import errno
 import socket
 import struct
+import json
 from os import environ
 
-try:
-    import rethinkdb.pbcpp
-    protobuf_implementation = 'cpp'
-except ImportError:
-    protobuf_implementation = 'python'
-
 from rethinkdb import ql2_pb2 as p
 
+pResponse = p.Response.ResponseType
+pQuery = p.Query.QueryType
+
 from rethinkdb import repl # For the repl connection
 from rethinkdb.errors import *
-from rethinkdb.ast import Datum, DB, expr
+from rethinkdb.ast import RqlQuery, DB, recursively_convert_pseudotypes
+
+class Query(object):
+    def __init__(self, type, token, term, global_optargs):
+        self.type = type
+        self.token = token
+        self.term = term
+        self.global_optargs = global_optargs
+
+    def serialize(self):
+        res = [self.type]
+        if self.term is not None:
+            res.append(self.term.build())
+        if self.global_optargs is not None:
+            optargs = { }
+            for (k,v) in self.global_optargs.iteritems():
+                optargs[k] = v.build() if isinstance(v, RqlQuery) else v
+            res.append(optargs)
+        return json.dumps(res, ensure_ascii=False, allow_nan=False)
+
+class Response(object):
+    def __init__(self, token, json_str):
+        json_str = json_str.decode('utf-8')
+        self.token = token
+        full_response = json.loads(json_str)
+        self.type = full_response["t"]
+        self.data = full_response["r"]
+        self.backtrace = full_response.get("b", None)
+        self.profile = full_response.get("p", None)
 
 class Cursor(object):
-    def __init__(self, conn, query, term, format_opts, opts):
+    def __init__(self, conn, query, opts):
         self.conn = conn
         self.query = query
-        self.term = term
-        self.format_opts = format_opts
         self.opts = opts
         self.responses = [ ]
         self.outstanding_requests = 0
@@ -32,15 +56,14 @@ class Cursor(object):
         self.connection_closed = False
 
     def _extend(self, response):
-        self.end_flag = response.type != p.Response.SUCCESS_PARTIAL
+        self.end_flag = response.type != pResponse.SUCCESS_PARTIAL and \
+                        response.type != pResponse.SUCCESS_FEED
         self.responses.append(response)
 
         if len(self.responses) == 1 and not self.end_flag:
             self.conn._async_continue_cursor(self)
 
     def __iter__(self):
-        format_opts = self.format_opts
-        deconstruct = Datum.deconstruct
         while True:
             if len(self.responses) == 0 and self.connection_closed:
                 raise RqlDriverError("Connection closed, cannot read cursor")
@@ -52,13 +75,16 @@ class Cursor(object):
             if len(self.responses) == 0 and self.end_flag:
                 break
 
-            self.conn._check_error_response(self.responses[0], self.term)
-            if self.responses[0].type != p.Response.SUCCESS_PARTIAL and self.responses[0].type != p.Response.SUCCESS_SEQUENCE:
+            self.conn._check_error_response(self.responses[0], self.query.term)
+            if self.responses[0].type != pResponse.SUCCESS_PARTIAL and \
+               self.responses[0].type != pResponse.SUCCESS_SEQUENCE and \
+               self.responses[0].type != pResponse.SUCCESS_FEED:
                 raise RqlDriverError("Unexpected response type received for cursor")
 
-            for datum in self.responses[0].response:
-                yield deconstruct(datum, format_opts)
+            response_data = recursively_convert_pseudotypes(self.responses[0].data, self.opts)
             del self.responses[0]
+            for item in response_data:
+                yield item
 
     def close(self):
         if not self.end_flag:
@@ -100,8 +126,9 @@ class Connection(object):
         except Exception as err:
             raise RqlDriverError("Could not connect to %s:%s. Error: %s" % (self.host, self.port, err))
 
-        self._sock_sendall(struct.pack("<L", p.VersionDummy.V0_2))
+        self._sock_sendall(struct.pack("<L", p.VersionDummy.Version.V0_3))
         self._sock_sendall(struct.pack("<L", len(self.auth_key)) + str.encode(self.auth_key, 'ascii'))
+        self._sock_sendall(struct.pack("<L", p.VersionDummy.Protocol.JSON))
 
         # Read out the response from the server, which will be a null-terminated string
         response = b""
@@ -121,7 +148,7 @@ class Connection(object):
         self.socket.settimeout(None)
 
     def close(self, noreply_wait=True):
-        if self.socket:
+        if self.socket is not None:
             if noreply_wait:
                 self.noreply_wait()
             try:
@@ -140,12 +167,10 @@ class Connection(object):
         self.next_token += 1
 
         # Construct query
-        query = p.Query()
-        query.type = p.Query.NOREPLY_WAIT
-        query.token = token
+        query = Query(pQuery.NOREPLY_WAIT, token, None, None)
 
         # Send the request
-        return self._send_query(query, 'noreply_wait')
+        return self._send_query(query)
 
     # Not thread safe. Sets this connection as global state that will be used
     # by subsequence calls to `query.run`. Useful for trying out RethinkDB in
@@ -170,40 +195,30 @@ class Connection(object):
                 if e.errno != errno.EINTR:
                     raise
 
-    def _start(self, term, **global_opt_args):
-        token = self.next_token
-        self.next_token += 1
-
-        # Construct query
-        query = p.Query()
-        query.type = p.Query.START
-        query.token = token
-
+    def _start(self, term, **global_optargs):
         # Set global opt args
-
         # The 'db' option will default to this connection's default
         # if not otherwise specified.
-        if 'db' in global_opt_args:
-            global_opt_args['db'] = DB(global_opt_args['db'])
+        if 'db' in global_optargs:
+            global_optargs['db'] = DB(global_optargs['db'])
         else:
             if self.db:
-               global_opt_args['db'] = DB(self.db)
+               global_optargs['db'] = DB(self.db)
 
-        for k,v in global_opt_args.items():
-            pair = query.global_optargs.add()
-            pair.key = k
-            expr(v).build(pair.val)
+        # Construct query
+        query = Query(pQuery.START, self.next_token, term, global_optargs)
+        self.next_token += 1
 
-        # Compile query to protobuf
-        term.build(query.query)
-        return self._send_query(query, term, global_opt_args)
+        return self._send_query(query, global_optargs)
 
     def _handle_cursor_response(self, response):
         cursor = self.cursor_cache[response.token]
         cursor._extend(response)
         cursor.outstanding_requests -= 1
 
-        if response.type != p.Response.SUCCESS_PARTIAL and cursor.outstanding_requests == 0:
+        if response.type != pResponse.SUCCESS_PARTIAL and \
+           response.type != pResponse.SUCCESS_FEED and \
+           cursor.outstanding_requests == 0:
             del self.cursor_cache[response.token]
 
     def _continue_cursor(self, cursor):
@@ -215,18 +230,14 @@ class Connection(object):
             return
 
         cursor.outstanding_requests = 1
-        query = p.Query()
-        query.type = p.Query.CONTINUE
-        query.token = cursor.query.token
-        self._send_query(query, cursor.term, cursor.opts, async=True)
+        query = Query(pQuery.CONTINUE, cursor.query.token, None, None)
+        self._send_query(query, cursor.opts, async=True)
 
     def _end_cursor(self, cursor):
         self.cursor_cache[cursor.query.token].outstanding_requests += 1
 
-        query = p.Query()
-        query.type = p.Query.STOP
-        query.token = cursor.query.token
-        self._send_query(cursor.query, cursor.term, async=True)
+        query = Query(pQuery.STOP, cursor.query.token, None, None)
+        self._send_query(query, async=True)
         self._handle_cursor_response(self._read_response(cursor.query.token))
 
     def _read_response(self, token):
@@ -235,14 +246,15 @@ class Connection(object):
             response_buf = b''
             try:
                 response_header = b''
-                while len(response_header) < 4:
-                    chunk = self._sock_recv(4)
+                while len(response_header) < 12:
+                    chunk = self._sock_recv(12 - len(response_header))
                     if len(chunk) == 0:
                         raise RqlDriverError("Connection is closed.")
                     response_header += chunk
 
-                # The first 4 bytes give the expected length of this response
-                (response_len,) = struct.unpack("<L", response_header)
+                # The first 8 bytes given the corresponding query token of this response
+                # The next 4 bytes give the expected length of this response
+                (response_token,response_len,) = struct.unpack("<qL", response_header)
 
                 while len(response_buf) < response_len:
                     chunk = self._sock_recv(response_len - len(response_buf))
@@ -256,8 +268,7 @@ class Connection(object):
                 raise err
 
             # Construct response
-            response = p.Response()
-            response.ParseFromString(response_buf)
+            response = Response(response_token, response_buf)
 
             # Check that this is the response we were expecting
             if response.token == token:
@@ -269,78 +280,61 @@ class Connection(object):
                 raise RqlDriverError("Unexpected response received.")
 
     def _check_error_response(self, response, term):
-        if response.type == p.Response.RUNTIME_ERROR:
-            message = Datum.deconstruct(response.response[0])
-            backtrace = response.backtrace
-            frames = backtrace.frames or []
+        if response.type == pResponse.RUNTIME_ERROR:
+            message = response.data[0]
+            frames = response.backtrace
             raise RqlRuntimeError(message, term, frames)
-        elif response.type == p.Response.COMPILE_ERROR:
-            message = Datum.deconstruct(response.response[0])
-            backtrace = response.backtrace
-            frames = backtrace.frames or []
+        elif response.type == pResponse.COMPILE_ERROR:
+            message = response.data[0]
+            frames = response.backtrace
             raise RqlCompileError(message, term, frames)
-        elif response.type == p.Response.CLIENT_ERROR:
-            message = Datum.deconstruct(response.response[0])
-            backtrace = response.backtrace
-            frames = backtrace.frames or []
+        elif response.type == pResponse.CLIENT_ERROR:
+            message = response.data[0]
+            frames = response.backtrace
             raise RqlClientError(message, term, frames)
 
-    def _send_query(self, query, term, opts={}, async=False):
+    def _send_query(self, query, opts={}, async=False):
         # Error if this connection has closed
-        if not self.socket:
+        if self.socket is None:
             raise RqlDriverError("Connection is closed.")
 
         query.accepts_r_json = True
 
-        # Send protobuf
-        query_protobuf = query.SerializeToString()
-        query_header = struct.pack("<L", len(query_protobuf))
-        self._sock_sendall(query_header + query_protobuf)
+        # Send json
+        query_str = query.serialize().encode('utf-8')
+        query_header = struct.pack("<QL", query.token, len(query_str))
+        self._sock_sendall(query_header + query_str)
 
-        if 'noreply' in opts and opts['noreply']:
-            return None
-        elif async:
+        if async or ('noreply' in opts and opts['noreply']):
             return None
 
         # Get response
         response = self._read_response(query.token)
+        self._check_error_response(response, query.term)
 
-        self._check_error_response(response, term)
-
-        format_opts = {}
-        if 'time_format' in opts:
-            format_opts['time_format'] = opts['time_format']
-        if 'group_format' in opts:
-            format_opts['group_format'] = opts['group_format']
-
-        # Sequence responses
-        if response.type == p.Response.SUCCESS_PARTIAL or response.type == p.Response.SUCCESS_SEQUENCE:
-            value = Cursor(self, query, term, format_opts, opts)
+        if response.type == pResponse.SUCCESS_PARTIAL or \
+           response.type == pResponse.SUCCESS_SEQUENCE or \
+           response.type == pResponse.SUCCESS_FEED:
+            # Sequence responses
+            value = Cursor(self, query, opts)
             self.cursor_cache[query.token] = value
             value._extend(response)
-
-        # Atom response
-        elif response.type == p.Response.SUCCESS_ATOM:
-            if len(response.response) < 1:
+        elif response.type == pResponse.SUCCESS_ATOM:
+            # Atom response
+            if len(response.data) < 1:
                 value = None
-            value = Datum.deconstruct(response.response[0], format_opts)
-
-        # Noreply_wait response
-        elif response.type == p.Response.WAIT_COMPLETE:
+            value = recursively_convert_pseudotypes(response.data[0], opts)
+        elif response.type == pResponse.WAIT_COMPLETE:
+            # Noreply_wait response
             return None
-
-        # Default for unknown response types
         else:
+            # Default for unknown response types
             raise RqlDriverError("Unknown Response type %d encountered in response." % response.type)
 
-        try:
-            if  Datum.deconstruct(response.profile) == None:
-                return value
-            else:
-                return {"value": value, "profile": Datum.deconstruct(response.profile)}
-        except AttributeError:
-            # response.profile does not exist
-            return value
+        if response.profile is not None:
+            value = {"value": value, "profile": response.profile}
+
+        return value
 
 def connect(host='localhost', port=28015, db=None, auth_key="", timeout=20):
     return Connection(host, port, db, auth_key, timeout)
